@@ -1,5 +1,4 @@
 int xleft;			/* the first visible column */
-int xquit;			/* exit if positive, force quit if negative */
 int xvis;			/* visual mode */
 int xai = 1;			/* autoindent option */
 int xic = 1;			/* ignorecase option */
@@ -21,7 +20,10 @@ int xpr;			/* ex_cprint register */
 int xsep = ':';			/* ex command separator */
 int xlim = -1;			/* rendering cutoff for non cursor lines */
 int xseq = 1;			/* undo/redo sequence */
+int xerr = 1;			/* error handling -
+				bit 1: print errors, bit 2: early return, bit 3: ignore errors */
 
+int xquit;			/* exit if positive, force quit if negative */
 int xrow, xoff, xtop;		/* current row, column, and top row */
 int xbufcur;			/* number of active buffers */
 int xgrec;			/* global vi/ex recursion depth */
@@ -132,12 +134,14 @@ void temp_pos(int i, int row, int off, int top)
 	tempbufs[i].top = top;
 }
 
-void temp_switch(int i)
+void temp_switch(int i, int swap)
 {
 	if (ex_buf == &tempbufs[i]) {
-		exbuf_save(ex_buf)
-		ex_buf = ex_pbuf;
-		ex_pbuf = ex_tpbuf;
+		if (swap) {
+			exbuf_save(ex_buf)
+			ex_buf = ex_pbuf;
+			ex_pbuf = ex_tpbuf;
+		}
 	} else {
 		if (!istempbuf(ex_buf)) {
 			ex_tpbuf = ex_pbuf;
@@ -191,7 +195,7 @@ static char *ex_reread(char ***re, int *dir)
 	return NULL;
 }
 
-static int ex_range(char **num, int n, int *row)
+static int ex_range(char *ploc, char **num, int n, int *row)
 {
 	int dir, off, beg, end, adj = 0;
 	while (**num == ' ' || **num == '\t')
@@ -200,6 +204,9 @@ static int ex_range(char **num, int n, int *row)
 	case '.':
 		++*num;
 		break;
+	case '%':
+		if (ploc != *num)
+			break;
 	case '$':
 		n = row ? lbuf_eol(xb, *row, 2) : lbuf_len(xb) - 1;
 		++*num;
@@ -219,6 +226,8 @@ static int ex_range(char **num, int n, int *row)
 		}
 		beg = row ? *row : xrow + (xkwddir > 0);
 		end = row ? beg+1 : lbuf_len(xb);
+		if (beg < 0 || beg >= lbuf_len(xb))
+			return -1;
 		if (lbuf_search(xb, xkwdrs, xkwddir, &beg,
 				&off, end, MIN(dir, 0))) {
 			xrerr = xsrerr;
@@ -236,16 +245,16 @@ static int ex_range(char **num, int n, int *row)
 	}
 	while (**num) {
 		dir = atoi(*num+1);
-		if (**num == '-') 
+		if (**num == '-')
 			n -= dir;
 		else if (**num == '+')
 			n += dir;
-		else if (**num == '/')
-			n /= dir ? dir : 1;
 		else if (**num == '*')
 			n *= dir;
-		else if (**num == '%')
-			n %= dir ? dir : 1;
+		else if (**num == '/' && dir)
+			n /= dir;
+		else if (**num == '%' && dir)
+			n %= dir;
 		else
 			break;
 		for (++*num; isdigit((unsigned char)**num);)
@@ -258,136 +267,81 @@ static int ex_range(char **num, int n, int *row)
 #define ex_vregion(loc, beg, end) ex_region(loc, beg, end, &xoff, &xoff)
 static int ex_region(char *loc, int *beg, int *end, int *o1, int *o2)
 {
-	if (!strcmp("%", loc) || !lbuf_len(xb)) {
-		*beg = 0;
-		*end = MAX(0, lbuf_len(xb));
-		return 0;
-	}
-	int vaddr = 0, haddr = 0, ooff = xoff, row;
+	int vaddr = *loc == '%', haddr = 0, update = 0, row = xrow, ooff = xoff;
+	char *ploc = loc;
 	xrerr = xirerr;
+	if (vaddr)
+		*beg = 0;
 	while (*loc) {
-		if (*loc == ';') {
-			loc++;
-			row = vaddr ? vaddr % 2 ? *beg : *end-1 : xrow;
-			if ((ooff = ex_range(&loc, ooff, &row)) < 0)
+		if (*loc == '|') {
+			char *cmd = ++loc;
+			for (; *loc && *loc != '|'; loc++)
+				if (*loc == '\\' && loc[1] == '|')
+					loc++;
+			int nulled = *loc;
+			*loc = '\0';
+			void *err = ex_exec(cmd);
+			if (nulled)
+				*loc++ = '|';
+			if (err) {
+				xrerr = "ex command error";
+				return 1;
+			}
+			continue;
+		} else if (*loc == ';') {
+			update = loc[1] == '#';
+			loc += 1 + update;
+			if ((ooff = ex_range(ploc, &loc, update ? ooff : xoff, &row)) < 0)
 				return 1;
 			if (haddr++ % 2)
 				*o2 = ooff;
 			else
 				*o1 = ooff;
 		} else {
-			if (*loc == ',')
-				loc++;
-			if (vaddr++ % 2) {
-				*end = ex_range(&loc, xrow, NULL) + 1;
-				if (*end <= *beg || *end > lbuf_len(xb))
-					return 1;
-			} else {
-				*beg = ex_range(&loc, xrow, NULL);
-				if (*beg < 0 || *beg >= lbuf_len(xb))
-					return 1;
+			if (*loc == ',') {
+				update = loc[1] == '#';
+				loc += 1 + update;
 			}
+			row = ex_range(ploc, &loc, update ? row : xrow, NULL);
+			if (vaddr++ % 2)
+				*end = row + 1;
+			else
+				*beg = row;
 		}
-		while (*loc && *loc != ';' && *loc != ',')
+		while (*loc && *loc != '|' && *loc != ';' && *loc != ',')
 		        loc++;
 	}
 	if (!vaddr) {
 		*beg = xrow;
-		*end = MIN(lbuf_len(xb), xrow + 1);
+		*end = MIN(lbuf_len(xb), *beg + 1);
 	} else if (vaddr == 1)
 		*end = *beg + 1;
-	return 0;
+	return *beg < 0 || *beg >= lbuf_len(xb) || *end <= *beg || *end > lbuf_len(xb);
 }
 
-static void *ec_find(char *loc, char *cmd, char *arg)
+static int ex_read(sbuf *sb, char *msg, char *ft, ins_state *is, int ps, int flg)
 {
-	int dir, off, obeg, beg, end;
-	char **s = &arg;
-	char *err = ex_reread(&s, &dir);
-	if (err)
-		return err;
-	if (ex_vregion(loc, &beg, &end))
-		return xrerr;
-	off = xoff;
-	obeg = beg;
-	if (xrow < beg || xrow > end) {
-		off = 0;
-		beg = xkwddir > 0 ? beg : end++;
-	} else
-		beg = xrow;
-	if (lbuf_search(xb, xkwdrs, xkwddir, &beg,
-			&off, end, cmd[1] == '+' ? xkwddir : MIN(dir, 0)))
-		return xuerr;
-	if (beg < obeg)
-		return xuerr;
-	xrow = beg;
-	xoff = off;
-	return NULL;
-}
-
-static void *ec_buffer(char *loc, char *cmd, char *arg)
-{
-	if (!arg[0]) {
-		char ln[512];
-		for (int i = 0; i < xbufcur; i++) {
-			char c = ex_buf == bufs+i ? '%' : ' ';
-			c = ex_pbuf == bufs+i ? '#' : c;
-			snprintf(ln, LEN(ln), "%d %c %s", i,
-				c + (char)bufs[i].lb->modified, bufs[i].path);
-			ex_print(ln, msg_ft)
+	int n = sb->s_n, key;
+	if (xvis & 2) {
+		while ((key = term_read()) != '\n') {
+			sbuf_chr(sb, key)
+			if (flg & 2 || xquit)
+				break;
 		}
-		return NULL;
-	} else if (atoi(arg) < 0) {
-		if (abs(atoi(arg)) <= LEN(tempbufs)) {
-			temp_switch(abs(atoi(arg))-1);
-			return NULL;
-		}
-	} else if (atoi(arg) < xbufcur) {
-		bufs_switchwft(atoi(arg))
-		return NULL;
+		sbuf_null(sb)
+		return key;
 	}
-	return "no such buffer";
-}
-
-static void *ec_quit(char *loc, char *cmd, char *arg)
-{
-	for (int i = 0; !strchr(cmd, '!') && i < xbufcur; i++)
-		if ((xquit < 0 || xgrec < 2) && bufs[i].lb->modified)
-			return "buffers modified";
-	if (!xquit)
-		xquit = !strchr(cmd, '!') ? 1 : -1;
-	return NULL;
-}
-
-void ex_bufpostfix(struct buf *p, int clear)
-{
-	p->mtime = mtime(p->path);
-	p->ft = syn_filetype(p->path);
-	lbuf_saved(p->lb, clear);
-}
-
-static int ex_read(sbuf *sb, char *msg, char *ft, int ps, int hist)
-{
-	if (!(xvis & 2)) {
-		preserve(int, xleft, xleft = 0;)
-		int n = sb->s_n, key;
-		sbuf_str(sb, msg)
+	preserve(int, xleft, xleft = 0;)
+	sbuf_str(sb, msg)
+	if (ft)
 		syn_setft(ft);
-		led_prompt(sb, NULL, &xkmap, &key, ps, hist);
+	key = led_prompt(sb, NULL, &xkmap, is, ps, flg);
+	if (ft)
 		syn_setft(xb_ft);
-		restore(xleft)
-		if (key != '\n')
-			return 1;
-		else if (!*msg || strcmp(sb->s + n, msg))
-			term_chr('\n');
-		return 0;
-	}
-	for (int c; !xquit && (c = term_read()) != '\n';)
-		sbuf_chr(sb, c)
-	if (xquit)
-		return 1;
-	sbuf_null(sb)
-	return 0;
+	restore(xleft)
+	if (key == '\n' && (!*msg || strcmp(sb->s + n, msg)))
+		term_chr('\n');
+	return key;
 }
 
 #define readfile(errchk) \
@@ -440,68 +394,185 @@ static void *ec_edit(char *loc, char *cmd, char *arg)
 			fd < 0 || rd ? 'f' : 'r');
 	if (!(xvis & 8))
 		ex_print(msg, bar_ft)
-	return fd < 0 || rd ? xuerr : NULL;
+	return (fd < 0 || rd) && *arg ? xuerr : NULL;
 }
 
-static void *ec_editapprox(char *loc, char *cmd, char *arg)
+static void *ec_fuzz(char *loc, char *cmd, char *arg)
 {
+	rset *rs;
+	char *path, *p, buf[32], trunc[100];
+	int c, pos, subs[2], inst = -1, lnum = -1;
+	int beg, end, max = INT_MAX;
+	int flg = REG_NEWLINE | REG_NOCAP;
+	ins_state is;
+	ins_init(is)
+	if (*cmd !='f')
+		temp_switch(1, 0);
+	if (ex_vregion(loc, &beg, &end)) {
+		if (*cmd !='f')
+			temp_switch(1, 1);
+		return xrerr;
+	}
+	if (!*loc) {
+		beg = 0;
+		end = lbuf_len(xb);
+		max = xrows * 3;
+	}
+	snprintf(trunc, sizeof(trunc), "truncated to %d lines", max);
 	sbuf_smake(sb, 128)
 	sbuf_smake(fuzz, 16)
-	char *path, *p, buf[32];
-	struct lbuf *lb = tempbufs[1].lb;
-	int z, c, pos, i, inst = *loc ? atoi(loc) : -1;
+	sbuf_smake(cmdbuf, 16)
 	sbuf_str(fuzz, arg)
-	for (z = 0;; z++) {
+	syn_setft(msg_ft);
+	while(1) {
 		sbuf_null(fuzz)
-		for (c = 0, pos = 0; pos < lbuf_len(lb); pos++) {
-			path = lb->ln[pos];
-			for (i = lbuf_s(path)->len; i > 0 && path[i] != '/'; i--);
-			if (strstr(&path[i+!!i], arg)) {
-				sbuf_mem(sb, &path, (int)sizeof(path))
-				p = itoa(c++, buf);
-				RS(2, ex_cprint(buf, msg_ft, -1, 0, 0))
-				RS(2, ex_cprint(path, msg_ft, -1, xleft ? 0 : (p - buf) + 1, 1))
+		c = 0;
+		rs = rset_smake(fuzz->s, xic ? flg | REG_ICASE : flg);
+		if (rs) {
+			syn_reloadft(syn_addhl(fuzz->s, 1), rs->regex->flg);
+			term_record = !!term_sbuf;
+			end = MIN(end, lbuf_len(xb));
+			for (pos = beg; c < max && pos < end; pos++) {
+				path = xb->ln[pos];
+				if (rset_match(rs, path, 0)) {
+					sbuf_mem(sb, &pos, (int)sizeof(pos))
+					p = itoa(c++, buf);
+					ex_cprint2(buf, NULL, -1, 0, 0, 2)
+					ex_cprint2(path, NULL, -1, (p - buf) + 1, 0, 3)
+				}
+			}
+			if (c == max && c != end)
+				ex_cprint2(trunc, NULL, -1, 0, 0, 3)
+			if (term_record)
+				term_commit();
+		}
+		if ((inst = ex_read(fuzz, "", NULL, &is, 0, 2)) == '\n' && c) {
+			if (c == 1)
+				break;
+			if ((inst = ex_read(cmdbuf, "", NULL, NULL, 0, 0)) == '\n') {
+				inst = atoi(cmdbuf->s);
+				break;
 			}
 		}
-		if ((inst < 0 && c > 1) || (z && !c)) {
-			RS(2, ex_cprint(fuzz->s, msg_ft, -1, 0, 0))
-			inst = term_read();
-			if (TK_INT(inst))
-				goto out;
-			else if (c && inst == '\n') {
-				inst = 0;
-				sbuf_cut(fuzz, inst)
-				if (!ex_read(fuzz, "", msg_ft, inst, inst))
-					inst = atoi(fuzz->s);	
+		if (TK_INT(inst))
+			goto ret;
+		if (c && c < 11 && isdigit(inst)) {
+			inst -= '0';
+			if (inst < c) {
+				fuzz->s_n--;
 				break;
-			} else if (c && c < 10 && isdigit(inst)) {
-				inst -= '0';
-				break;
-			} else if (inst == 127 || inst == TK_CTL('h'))
-				sbuf_cut(fuzz, MAX(0, fuzz->s_n-1))
-			else
-				sbuf_chr(fuzz, inst)
-			sbuf_cut(sb, 0)
-			arg = fuzz->s;
-			if (c)
-				ex_print("", msg_ft)
-		} else
-			break;
-		inst = -1;
+			}
+		}
+		rset_free(rs);
+		sbuf_cut(sb, 0)
+		if (c && xvis & 4)
+			ex_print("", NULL)
+		else if (c) {
+			term_clean();
+			term_pos(xrows, 0);
+		}
 	}
-	if ((inst >= 0 && inst < c) || c == 1) {
-		path = *((char**)sb->s + (c == 1 ? 0 : inst));
+	if ((inst >= 0 && inst < c) || c == 1)
+		lnum = *((int*)sb->s + (c == 1 ? 0 : inst));
+	ret:
+	syn_setft(xb_ft);
+	if (fuzz->s_n) {
+		sbuf_cut(cmdbuf, 0)
+		sbuf_str(cmdbuf, loc)
+		sbuf_str(cmdbuf, cmd)
+		sbuf_chr(cmdbuf, ' ')
+		sbufn_mem(cmdbuf, fuzz->s, fuzz->s_n)
+		lbuf_dedup(tempbufs[0].lb, cmdbuf->s, cmdbuf->s_n)
+		temp_pos(0, -1, 0, 0);
+		temp_write(0, cmdbuf->s);
+	}
+	free(cmdbuf->s);
+	free(fuzz->s);
+	free(sb->s);
+	path = lbuf_get(xb, lnum);
+	if (*cmd == 'f' && path) {
+		rset_find(rs, path, subs, 0);
+		xrow = lnum;
+		xoff = subs[0];
+	} else if (path) {
 		path[lbuf_s(path)->len] = '\0';
 		ec_edit(loc, cmd, path);
 		path[lbuf_s(path)->len] = '\n';
-	}
-	out:
-	xmpt = xmpt >= 0 ? 0 : xmpt;
-	free(fuzz->s);
-	free(sb->s);
+	} else if (*cmd != 'f')
+		temp_switch(1, 1);
+	rset_free(rs);
 	return NULL;
 }
 
+static void *ec_find(char *loc, char *cmd, char *arg)
+{
+	int dir, off, obeg, beg, end;
+	if (!*arg || (*arg != '>' && *arg != '<'))
+		return ec_fuzz(loc, cmd, arg + (*arg == '\\' && arg[1]));
+	char **s = &arg;
+	char *err = ex_reread(&s, &dir);
+	if (err)
+		return err;
+	if (ex_vregion(loc, &beg, &end))
+		return xrerr;
+	off = xoff;
+	obeg = beg;
+	if (xrow < beg || xrow > end) {
+		off = 0;
+		beg = xkwddir > 0 ? beg : end++;
+	} else
+		beg = xrow;
+	if (lbuf_search(xb, xkwdrs, xkwddir, &beg,
+			&off, end, cmd[1] == '+' ? xkwddir : MIN(dir, 0)))
+		return xuerr;
+	if (beg < obeg)
+		return xuerr;
+	xrow = beg;
+	xoff = off;
+	return NULL;
+}
+
+static void *ec_buffer(char *loc, char *cmd, char *arg)
+{
+	if (!arg[0]) {
+		char ln[512];
+		for (int i = 0; i < xbufcur; i++) {
+			char c = ex_buf == bufs+i ? '%' : ' ';
+			c = ex_pbuf == bufs+i ? '#' : c;
+			snprintf(ln, LEN(ln), "%d %c %s", i,
+				c + (char)bufs[i].lb->modified, bufs[i].path);
+			ex_print(ln, msg_ft)
+		}
+		return NULL;
+	} else if (atoi(arg) < 0) {
+		if (abs(atoi(arg)) <= LEN(tempbufs)) {
+			temp_switch(abs(atoi(arg))-1, 1);
+			return NULL;
+		}
+	} else if (atoi(arg) < xbufcur) {
+		bufs_switchwft(atoi(arg))
+		return NULL;
+	}
+	return "no such buffer";
+}
+
+static void *ec_quit(char *loc, char *cmd, char *arg)
+{
+	for (int i = 0; !strchr(cmd, '!') && i < xbufcur; i++)
+		if ((xquit < 0 || xgrec < 2) && bufs[i].lb->modified)
+			return "buffers modified";
+	if (!xquit)
+		xquit = !strchr(cmd, '!') ? 1 : -1;
+	return NULL;
+}
+
+void ex_bufpostfix(struct buf *p, int clear)
+{
+	p->mtime = mtime(p->path);
+	p->ft = syn_filetype(p->path);
+	lbuf_saved(p->lb, clear);
+}
+
 static void *ec_setpath(char *loc, char *cmd, char *arg)
 {
 	free(xb_path);
@@ -515,7 +586,7 @@ static void *ec_read(char *loc, char *cmd, char *arg)
 	sbuf obuf, *sb;
 	char msg[512];
 	char *path, *ret = NULL;
-	int beg, end, o1 = 0, o2 = -1;
+	int beg = 0, end = 0, o1 = 0, o2 = -1;
 	int row = xrow, off = xoff, fd = -1;
 	struct lbuf *lb = lbuf_make(), *pxb = xb;
 	path = arg[0] ? arg : xb_path;
@@ -537,10 +608,10 @@ static void *ec_read(char *loc, char *cmd, char *arg)
 	xb = lb;
 	xrow = 0;
 	xoff = 0;
-	if (ex_region(loc, &beg, &end, &o1, &o2)) {
+	if (lbuf_len(lb) && ex_region(loc, &beg, &end, &o1, &o2)) {
 		ret = xrerr;
 		goto err;
-	} else if (!loc[0]) {
+	} else if (!*loc) {
 		beg = 0;
 		end = lbuf_len(lb);
 	}
@@ -575,13 +646,13 @@ static void *ec_write(char *loc, char *cmd, char *arg)
 {
 	char msg[512], *path;
 	sbuf ibuf;
-	int fd, beg, end, o1 = -1, o2 = -1;
+	int fd, beg = 0, end = 0, o1 = -1, o2 = -1;
 	path = arg[0] ? arg : xb_path;
 	if (cmd[0] == 'x' && !xb->modified)
 		return ec_quit("", cmd, "");
-	if (ex_region(loc, &beg, &end, &o1, &o2))
+	if (lbuf_len(xb) && ex_region(loc, &beg, &end, &o1, &o2))
 		return xrerr;
-	if (!loc[0]) {
+	if (!*loc) {
 		beg = 0;
 		end = lbuf_len(xb);
 	}
@@ -628,45 +699,49 @@ static void *ec_termexec(char *loc, char *cmd, char *arg)
 	return NULL;
 }
 
-void ex_cprint(char *line, char *ft, int r, int c, int ln)
+void ex_cprint(char *line, char *ft, int r, int c, int left, int flg)
 {
-	syn_blockhl = -1;
-	if (!(xvis & 4)) {
-		if (xmpt == 1)
-			term_chr('\n');
-		if (isupper(xpr) && xmpt == 1 && !strchr(line, '\n'))
-			ex_regput(xpr, "\n", 1);
-		term_pos(xrows, led_pos(vi_msg, 0));
-		snprintf(vi_msg+c, sizeof(vi_msg)-c, "%s", line);
-	}
-	if (xpr)
+	int lntest;
+	if (xpr) {
 		ex_regput(xpr, line, 1);
+		if (flg & 1 && isupper(xpr) &&
+				xregs[xpr] && xregs[xpr]->s_n &&
+				xregs[xpr]->s[xregs[xpr]->s_n-1] != '\n')
+			ex_regput(xpr, "\n", 1);
+	}
 	if (xvis & 2) {
 		term_write(line, dstrlen(line, '\n'))
 		term_write("\n", 1)
 		return;
 	}
-	syn_setft(ft);
-	led_crender(line, r, c, xleft, xleft + xcols - c)
-	syn_setft(xb_ft);
-	if (ln && (xvis & 4 || xmpt > 0)) {
+	syn_blockhl = -1;
+	if (!(xvis & 4) && !(flg & 2)) {
+		term_pos(xrows, 0);
+		snprintf(vi_msg+c, sizeof(vi_msg)-c, "%s", line);
+		lntest = xmpt;
+		if (lntest == 1)
+			term_chr('\n');
+		xmpt += lntest >= 0 && (flg & 1 || lntest == 1);
+	} else
+		lntest = 1;
+	if (ft)
+		syn_setft(ft);
+	led_crender(line, r, c, left, left + xcols - c)
+	if (ft)
+		syn_setft(xb_ft);
+	if (flg & 1 && lntest > 0)
 		term_chr('\n');
-		if (isupper(xpr) && !strchr(line, '\n'))
-			ex_regput(xpr, "\n", 1);
-	}
-	xmpt += !(xvis & 4) && xmpt >= 0 && (ln || xmpt);
 }
 
 static void *ec_insert(char *loc, char *cmd, char *arg)
 {
-	int beg, end, o1 = -1, o2 = -1, ps = 0;
-	if (ex_region(loc, &beg, &end, &o1, &o2))
+	int key, beg = 0, end = 0, o1 = -1, o2 = -1, ps = 0;
+	if (lbuf_len(xb) && ex_region(loc, &beg, &end, &o1, &o2))
 		return xrerr;
 	sbuf_smake(sb, 128)
 	if (*arg)
 		term_push(arg, strlen(arg));
-	vi_insmov = 0;
-	while (!ex_read(sb, "", msg_ft, ps, 0)) {
+	while ((key = ex_read(sb, "", msg_ft, NULL, ps, 0)) == '\n') {
 		if (xvis & 2 && !strcmp(".", sb->s + ps)) {
 			sb->s_n--;
 			break;
@@ -674,9 +749,9 @@ static void *ec_insert(char *loc, char *cmd, char *arg)
 		sbuf_chr(sb, '\n')
 		ps = sb->s_n;
 	}
-	if (vi_insmov == TK_CTL('c'))
+	if (key == TK_CTL('c'))
 		goto ret;
-	if (vi_insmov == 127 && sb->s_n && sb->s[sb->s_n-1] == '\n')
+	if (key == 127 && sb->s_n && sb->s[sb->s_n-1] == '\n')
 		sb->s_n--;
 	sbuf_null(sb)
 	if (cmd[0] == 'a' && (beg + 1 <= lbuf_len(xb)))
@@ -686,11 +761,11 @@ static void *ec_insert(char *loc, char *cmd, char *arg)
 	if (o1 >= 0 && cmd[0] == 'c') {
 		if (!sb->s_n && o2 <= o1)
 			goto ret;
-		char *p = lbuf_joinsb(xb, beg, beg, sb, &o1, &o2);
+		char *p = lbuf_joinsb(xb, beg, end-1, sb, &o1, &o2);
 		o1 -= sb->s[0] == '\n';
 		free(sb->s);
 		sb->s = p;
-	} else if (!(xvis & 2) && vi_insmov != 127)
+	} else if (!(xvis & 2) && key != 127)
 		sbufn_chr(sb, '\n')
 	else if (!sb->s_n)
 		goto ret;
@@ -732,12 +807,10 @@ static void *ec_print(char *loc, char *cmd, char *arg)
 		else {
 			if (xvis & 4 && beg == end-1)
 				xleft = ren_position(ln)->pos[MIN(xoff, rstate->n)];
-			ex_cprint(ln, msg_ft, -1, 0, 1);
+			ex_cprint(ln, msg_ft, -1, 0, xleft, 1);
 			continue;
 		}
-		preserve(int, xleft, xleft = 0;)
-		ex_cprint(o, msg_ft, -1, 0, 1);
-		restore(xleft)
+		ex_cprint(o, msg_ft, -1, 0, 0, 1);
 		free(o);
 	}
 	rstate--;
@@ -747,10 +820,19 @@ static void *ec_print(char *loc, char *cmd, char *arg)
 
 static void *ec_delete(char *loc, char *cmd, char *arg)
 {
-	int beg, end;
-	if (ex_vregion(loc, &beg, &end) || !lbuf_len(xb))
+	int beg, end, o1 = -1, o2 = -1;
+	sbuf sb;
+	char *p = NULL;
+	if (ex_region(loc, &beg, &end, &o1, &o2))
 		return xrerr;
-	lbuf_edit(xb, NULL, beg, end, 0, 0);
+	if (o1 >= 0) {
+		sb.s = "";
+		sb.s_n = 0;
+		p = lbuf_joinsb(xb, beg, end-1, &sb, &o1, &o2);
+		xoff = o1;
+	}
+	lbuf_edit(xb, p, beg, end, o1, o2);
+	free(p);
 	xrow = beg;
 	return NULL;
 }
@@ -766,8 +848,7 @@ void ex_regput(unsigned char c, const char *s, int append)
 		if (!append)
 			sbuf_cut(sb, 0)
 		sbuf_str(sb, s)
-		sbuf_set(sb, '\0', 4)
-		sb->s_n -= 4;
+		sbufn_null(sb)
 	} else if (sb) {
 		sbuf_free(sb)
 		xregs[c] = NULL;
@@ -780,7 +861,7 @@ static void *ec_yank(char *loc, char *cmd, char *arg)
 	if (cmd[2] == '!') {
 		ex_regput(*arg, NULL, 0);
 		return NULL;
-	} else if (ex_region(loc, &beg, &end, &o1, &o2) || !lbuf_len(xb))
+	} else if (ex_region(loc, &beg, &end, &o1, &o2))
 		return xrerr;
 	sbuf sb;
 	lbuf_region(xb, &sb, beg, o1, end-1, o2);
@@ -791,7 +872,7 @@ static void *ec_yank(char *loc, char *cmd, char *arg)
 
 static void *ec_put(char *loc, char *cmd, char *arg)
 {
-	int beg, end, i = 0;
+	int beg = 0, end = 0, i = 0;
 	sbuf *buf;
 	if (!*arg || (arg[i] == '!' && arg[i+1] && arg[i+1] != ' '))
 		buf = xregs[i];
@@ -803,7 +884,7 @@ static void *ec_put(char *loc, char *cmd, char *arg)
 	if (arg[i] == '!' && arg[i+1])
 		return ex_pipeout(arg + i + 1, buf);
 	int n = lbuf_len(xb), o1 = -1, o2 = -1;
-	if (ex_region(loc, &beg, &end, &o1, &o2))
+	if (n && ex_region(loc, &beg, &end, &o1, &o2))
 		return xrerr;
 	if (o1 >= 0 && n) {
 		char *p = lbuf_joinsb(xb, end-1, end-1, buf, &o1, &o2);
@@ -875,7 +956,7 @@ static void *ec_substitute(char *loc, char *cmd, char *arg)
 		rep = re_read(&s);
 	}
 	free(pat);
-	int offs[rs->grpcnt * 2];
+	int offs[rs->nsubc];
 	for (i = beg; i < end; i++) {
 		char *ln = lbuf_get(xb, i);
 		sbuf *r = NULL;
@@ -894,7 +975,7 @@ static void *ec_substitute(char *loc, char *cmd, char *arg)
 					}
 					_rep++;
 					grp = abs((*_rep - '0') * 2);
-					if (grp + 1 >= rs->grpcnt * 2)
+					if (grp + 1 >= rs->nsubc)
 						sbuf_chr(r, (unsigned char)*_rep)
 					else if (offs[grp] >= 0)
 						sbuf_mem(r, ln + offs[grp], offs[grp + 1] - offs[grp])
@@ -909,7 +990,7 @@ static void *ec_substitute(char *loc, char *cmd, char *arg)
 		if (r) {
 			if (first < 0) {
 				first = i;
-				lo = lbuf_opt(xb, NULL, xrow, xoff, 0);
+				lo = lbuf_opt(xb, xrow, xoff, 0);
 				lbuf_smark(xb, lo, i, 0);
 				lbuf_emark(xb, lo, 0, 0);
 			}
@@ -920,7 +1001,7 @@ static void *ec_substitute(char *loc, char *cmd, char *arg)
 		}
 	}
 	if (first >= 0) {
-		lo = lbuf_opt(xb, NULL, xrow, xoff, 0);
+		lo = lbuf_opt(xb, xrow, xoff, 0);
 		lbuf_smark(xb, lo, first, 0);
 		lbuf_emark(xb, lo, last, 0);
 	}
@@ -932,10 +1013,10 @@ static void *ec_substitute(char *loc, char *cmd, char *arg)
 
 static void *ec_exec(char *loc, char *cmd, char *arg)
 {
-	int beg, end, o1 = 0, o2 = -1;
+	int beg = 0, end = 0, o1 = 0, o2 = -1;
 	if (!loc[0])
 		return ex_pipeout(arg, NULL);
-	if (ex_region(loc, &beg, &end, &o1, &o2))
+	if (lbuf_len(xb) && ex_region(loc, &beg, &end, &o1, &o2))
 		return xrerr;
 	sbuf text;
 	lbuf_region(xb, &text, beg, o1, end-1, o2);
@@ -962,7 +1043,7 @@ static void *ec_ft(char *loc, char *cmd, char *arg)
 		led_attsb = NULL;
 	}
 	for (int i = 1; i < 4; i++)
-		syn_reloadft(syn_findhl(i));
+		syn_reloadft(syn_findhl(i), 0);
 	return NULL;
 }
 
@@ -1001,7 +1082,7 @@ static void *ec_glob(char *loc, char *cmd, char *arg)
 	for (i = beg; i < lbuf_len(xb);) {
 		char *ln = lbuf_get(xb, i);
 		lbuf_s(ln)->grec &= ~xgdep;
-		if ((rset_find(rs, ln, NULL, REG_NEWLINE) < 0) == not) {
+		if (rset_match(rs, ln, REG_NEWLINE) != not) {
 			xrow = i;
 			if (ex_exec(s))
 				break;
@@ -1037,6 +1118,16 @@ static void *ec_while(char *loc, char *cmd, char *arg)
 	return !ret || addr == 1 ? NULL : memcpy(_skip, skip, sizeof(skip));
 }
 
+static void *ec_join(char *loc, char *cmd, char *arg)
+{
+	int beg, end, o2 = 0;
+	if (ex_vregion(loc, &beg, &end))
+		return xrerr;
+	lbuf_join(xb, beg, end+1, xoff, &o2, arg[0]);
+	xrow = beg;
+	return NULL;
+}
+
 static void *ec_setdir(char *loc, char *cmd, char *arg)
 {
 	free(fs_exdir);
@@ -1057,13 +1148,14 @@ static void *ec_chdir(char *loc, char *cmd, char *arg)
 	if (!getcwd(oldpath, sizeof(oldpath)))
 		if ((opath = getenv("PWD")))
 			strncpy(oldpath, opath, sizeof(oldpath)-1);
-	if (*arg && chdir(arg))
+	plen = strlen(oldpath);
+	i = plen == sizeof(oldpath)-1;
+	if (chdir(*arg ? arg : oldpath))
 		return "chdir error";
 	if (!getcwd(newpath, sizeof(newpath)))
 		return "getcwd error";
 	setenv("PWD", newpath, 1);
-	plen = strlen(oldpath);
-	if (plen == sizeof(oldpath)-1)
+	if (i)
 		return "oldpath >= 4096";
 	if (plen && oldpath[plen-1] != '/')
 		oldpath[plen++] = '/';
@@ -1137,8 +1229,8 @@ static void *ec_regprint(char *loc, char *cmd, char *arg)
 	for (int i = 1; i < LEN(xregs); i++) {
 		if (xregs[i] && i != xpr) {
 			*buf = i;
-			RS(2, ex_cprint(buf, msg_ft, -1, 0, 0))
-			RS(2, ex_cprint(xregs[i]->s, msg_ft, -1, xleft ? 0 : 2, 1))
+			ex_cprint2(buf, msg_ft, -1, 0, 0, 0)
+			ex_cprint2(xregs[i]->s, msg_ft, -1, xleft ? 0 : 2, xleft, 1)
 		}
 	}
 	restore(xtd)
@@ -1196,7 +1288,7 @@ static void *eo_##opt(char *loc, char *cmd, char *arg) { inner }
 
 EO(pac) EO(pr) EO(ai) EO(ish) EO(ic) EO(grp) EO(shape) EO(seq)
 EO(sep) EO(tbs) EO(td) EO(order) EO(hll) EO(hlw) EO(hlp) EO(hlr)
-EO(hl) EO(lim) EO(led) EO(vis) EO(mpt)
+EO(hl) EO(lim) EO(led) EO(vis) EO(mpt) EO(err)
 
 _EO(left,
 	if (*loc)
@@ -1230,8 +1322,9 @@ static struct excmd {
 	EO(ai),
 	{"ac", ec_setacreg},
 	{"a", ec_insert},
-	{"ea!", ec_editapprox},
-	{"ea", ec_editapprox},
+	EO(err),
+	{"ef!", ec_fuzz},
+	{"ef", ec_fuzz},
 	{"e!", ec_edit},
 	{"e", ec_edit},
 	{"ft", ec_ft},
@@ -1247,7 +1340,8 @@ static struct excmd {
 	EO(grp),
 	{"g!", ec_glob},
 	{"g", ec_glob},
-	{"k", ec_mark},
+	EO(mpt),
+	{"m", ec_mark},
 	{"q!", ec_quit},
 	{"q", ec_quit},
 	{"reg", ec_regprint},
@@ -1273,6 +1367,7 @@ static struct excmd {
 	{"cm", ec_cmap},
 	{"cd", ec_chdir},
 	{"c", ec_insert},
+	{"j", ec_join},
 	EO(tbs),
 	EO(td),
 	EO(order),
@@ -1285,7 +1380,6 @@ static struct excmd {
 	EO(lim),
 	EO(led),
 	EO(vis),
-	EO(mpt),
 	{"=", ec_num},
 	{"", ec_print}, /* do not remove */
 };
@@ -1296,22 +1390,26 @@ static const char *ex_arg(const char *src, sbuf *sb, int *arg)
 	*arg = sb->s_n;
 	while (*src && *src != xsep) {
 		if (*src == '%') {
-			int n = -1;
-			struct buf *pbuf = ex_buf;
 			src++;
-			if (*src == '#') {
-				src++;
-				pbuf = ex_pbuf;
-			} else if ((*src ^ '0') < 10)
-				pbuf = &bufs[n = atoi(src)];
-			if (pbuf < bufs || pbuf >= &bufs[xbufcur] || !pbuf->path[0]) {
-				ex_print("\"#\" or \"%\" is not set", msg_ft)
-				*arg = -1;
-			} else
-				sbuf_str(sb, pbuf->path)
-			if (n >= 0)
-				src += snprintf(0, 0, "%d", n);
-			src += *src == '\\' && src[-1] != '#' && (src[1] ^ '0') < 10;
+			if (*src == '@' && src[1] && src[1] != '\\') {
+				sbuf *reg = xregs[(unsigned char)src[1]];
+				if (reg)
+					sbuf_str(sb, reg->s)
+				src += 2;
+			} else {
+				struct buf *pbuf = ex_buf;
+				int n;
+				if (*src == '#') {
+					src++;
+					pbuf = ex_pbuf;
+				} else if ((*src ^ '0') < 10) {
+					pbuf = &bufs[n = atoi(src)];
+					src += snprintf(0, 0, "%d", n);
+				}
+				src += *src == '\\' && src[-1] != '#' && (src[1] ^ '0') < 10;
+				if (pbuf >= bufs && pbuf < &bufs[xbufcur] && pbuf->path[0])
+					sbuf_str(sb, pbuf->path)
+			}
 		} else if (*src == '!') {
 			int n = sb->s_n;
 			src++;
@@ -1346,10 +1444,10 @@ static const char *ex_cmd(const char *src, sbuf *sb, int *idx)
 	char *dst = sb->s;
 	while (*src == xsep || *src == ' ' || *src == '\t')
 		src += !!src[0];
-	while (memchr(" \t0123456789+-.,<>/$';%*", *src, 24)) {
+	while (memchr(" \t0123456789+-.,<>/$';%*#|", *src, 26)) {
 		if (*src == '\'' && src[1])
 			*dst++ = *src++;
-		if (*src == '>' || *src == '<') {
+		if (*src == '>' || *src == '<' || *src == '|') {
 			j = *src;
 			do {
 				*dst++ = *src++;
@@ -1385,21 +1483,25 @@ void *ex_exec(const char *ln)
 		lbuf_mark(xb, '*', xrow, xoff);
 	sbuf_smake(sb, strlen(ln) + 4)
 	for (int i = 0; *ln; i++) {
+		sbuf_cut(sb, 0)
 		ln = ex_arg(ex_cmd(ln, sb, &idx), sb, &arg);
-		if (i >= r1 && i <= r2)
-			continue;
-		if (arg >= 0) {
+		if (i < r1 || i > r2) {
 			ret = excmds[idx].ec(sb->s, excmds[idx].name, sb->s + arg);
-			if (ret && !*ret) {
+			if (!ret)
+				continue;
+			if (!*ret) {
 				r1 = ((int*)ret)[1] + i;
 				r2 = ((int*)ret)[2] + i;
-			} else if (ret && ret != xuerr)
+				continue;
+			}
+			if (ret != xuerr && xerr & 1)
 				ex_print(ret, msg_ft)
+			if (xerr & 2)
+				break;
 		}
-		sbuf_cut(sb, 0)
 	}
 	free(sb->s);
-	return ret;
+	return xerr & 4 ? NULL : ret;
 }
 
 /* ex main loop */
@@ -1409,7 +1511,7 @@ void ex(void)
 	int esc = 0;
 	sbuf_smake(sb, xcols)
 	while (!xquit) {
-		if (!ex_read(sb, ":", ex_ft, 0, 1)) {
+		if (ex_read(sb, ":", ex_ft, NULL, 0, 1) == '\n') {
 			if (!strcmp(sb->s, ":") && esc)
 				ex_exec(xregs[':']->s);
 			else
