@@ -24,7 +24,6 @@
 #include "uc.c"
 
 int vi_hidch;		/* show hidden chars */
-int vi_insmov;		/* moving in insert outside of insertion sbuf */
 int vi_lncol;		/* line numbers cursor offset */
 char vi_msg[512];	/* current message */
 static int vi_lnnum;	/* line numbers */
@@ -45,7 +44,6 @@ static int vi_scrolley;			/* scroll amount for ^e and ^y */
 static int vi_cndir = 1;		/* ^n direction */
 static int vi_status;			/* always show status */
 static int vi_tsm;			/* type of the status message */
-static int vi_joinmode = 1;		/* 1: insert extra space for pad 0: raw line join */
 static int vi_nlword;			/* new line mode for eEwWbB */
 
 void *emalloc(size_t size)
@@ -120,7 +118,7 @@ nrow = xrow; \
 noff = xoff; \
 for (i = 0, ret = 0;; i++) { \
 	l1 = ren_next(c, ren_pos(c, noff), 1)-1-xleft+vi_lncol; \
-	if (l1 > xcols || l1 < 0 || ret) \
+	if (l1 > xcols || l1 < 0 || ret || l1 >= rstate->cmax + vi_lncol) \
 		break; \
 	i = i > 99 ? i % 100 : i; \
 	itoa(i%10 ? i%10 : i, snum); \
@@ -133,7 +131,7 @@ static void vi_drawrow(int row)
 	int l1, i, i1, lnnum = vi_lnnum;
 	char *c, *s;
 	static char ch[5] = "~";
-	if (*vi_word) {
+	if (*vi_word && xled) {
 		int noff, nrow, ret;
 		s = lbuf_get(xb, row - vi_rshift);
 		c = lbuf_get(xb, xrow);
@@ -173,7 +171,7 @@ static void vi_drawrow(int row)
 	rstate += row != xrow;
 	if (!s)
 		s = row ? ch : ch+1;
-	else if (lnnum) {
+	else if (lnnum && xled) {
 		char tmp[32], tmp1[32], *p;
 		c = tmp, i = 0, i1 = 0;
 		if (lnnum == 1 || lnnum & 2) {
@@ -195,7 +193,7 @@ static void vi_drawrow(int row)
 		led_crender(s, row - xtop, l1, xleft, xleft + xcols - l1)
 		preserve(int, syn_blockhl, syn_blockhl = -1;)
 		syn_setft(nn_ft);
-		if ((lnnum == 1 || lnnum & 4) && xled && !xleft && vi_lncol) {
+		if ((lnnum == 1 || lnnum & 4) && !xleft && vi_lncol) {
 			for (i1 = 0; i1 < rstate->cmax &&
 					memchr(" \t", *rstate->chrs[ren_off(s, i1)], 2);)
 				i1 = ren_next(s, i1, 1);
@@ -260,7 +258,7 @@ static char *vi_prompt(char *msg, char *ft, char *insert, int *kmap, int *mlen)
 	*mlen = sb->s_n;
 	term_pos(xrows, led_pos(msg, 0));
 	syn_setft(ft);
-	led_prompt(sb, insert, kmap, &key, 0, 1);
+	key = led_prompt(sb, insert, kmap, NULL, 0, 1);
 	syn_setft(xb_ft);
 	strncpy(vi_msg, sb->s, sizeof(vi_msg) - 1);
 	if (key == '\n')
@@ -485,7 +483,7 @@ void dir_calc(char *path)
 				memcpy(cpath, ptrs[i], pathlen + len);
 				plen[i++] = pathlen + len;
 			} else if (ret >= 0 && S_ISREG(statbuf.st_mode))
-				if (!fsincl || rset_find(fsincl, cpath, NULL, 0) >= 0) {
+				if (!fsincl || rset_match(fsincl, cpath, 0)) {
 					sbuf_mem(sb, cpath, (int)(pathlen + len))
 					sbuf_chr(sb, '\n')
 				}
@@ -576,6 +574,7 @@ static void vc_status(int type)
 static int vi_motion(int vc, int *row, int *off)
 {
 	static sbuf *savepath[10];
+	static rset *bre;
 	static int srow[10], soff[10], lkwdcnt;
 	static int cadir = 1;
 	char *cs;
@@ -672,7 +671,8 @@ static int vi_motion(int vc, int *row, int *off)
 	case '(':
 	case ')':
 		dir = mv == '(' ? 1 : -1;
-		rset *set = rset_smake("^[.?!]+['\\])]*(?:[ \t]+\n?|\n)", 0);
+		if (!bre)
+			bre = rset_smake("^[.?!]+['\\])]*(?:[ \t]+\n?|\n)", 0);
 		int subs[2], org;
 		for (i = 0; i < cnt; i++) {
 			mark = *row;
@@ -691,7 +691,7 @@ static int vi_motion(int vc, int *row, int *off)
 						*row += 1;
 					*off = MAX(0, lbuf_indents(xb, *row));
 					break;
-				} else if (rset_find(set, cs, subs, 0) >= 0) {
+				} else if (rset_find(bre, cs, subs, 0) >= 0) {
 					if (mark == *row && rstate->chrs[org] == cs + subs[1])
 						continue;
 					if (!cs[subs[1]]) {
@@ -705,7 +705,6 @@ static int vi_motion(int vc, int *row, int *off)
 				}
 			}
 		}
-		rset_free(set);
 		return mv;
 	case '{':
 	case '}':
@@ -717,7 +716,7 @@ static int vi_motion(int vc, int *row, int *off)
 			if (lbuf_sectionbeg(xb, dir, row, off, mark))
 				break;
 		break;
-	case TK_CTL(']'):	/* note: this is also ^5 as per ascii */
+	case TK_CTL(']'):	/* this is also ^5 on some systems */
 	case TK_CTL('p'):
 		#define open_saved(n) \
 		if (savepath[n]) { \
@@ -857,38 +856,38 @@ static void vi_indents(char *ln, int *l)
 	*l = ln - pln;
 }
 
-static void vi_change(int r1, int o1, int r2, int o2, int lnmode)
+static int vi_change(int r1, int o1, int r2, int o2, int lnmode)
 {
-	char *post, *_post, *ln = lbuf_get(xb, r1);
+	char *post, *ln = lbuf_get(xb, r1);
 	sbuf rsb;
-	int tlen, l1, l2 = 1, postn = 1;
+	int key, tlen, l1, l2 = 1, postn = 1;
 	sbuf_smake(sb, xcols)
 	if (lnmode || !ln) {
 		vi_indents(ln, &l1);
 		o1 = l1;
-		post = _post = uc_dup("\n");
+		post = "\n";
 		tlen = -1;
 		lbuf_region(xb, &rsb, r1, 0, r2, -1);
 	} else {
 		l1 = uc_chr(ln, o1) - ln;
-		post = _post = uc_subl(lbuf_get(xb, r2), o2, -1, &l2, &postn);
+		post = uc_chr(lbuf_get(xb, r2), o2);
+		l2 = uc_chrn(post, -1, &postn) - post;
 		tlen = lbuf_s(ln)->len+1;
 		lbuf_region(xb, &rsb, r1, o1, r2, o2);
 	}
 	vi_regput(vi_ybuf, rsb.s, lnmode);
 	free(rsb.s);
 	term_pos(r1 - xtop < 0 ? 0 : r1 - xtop, 0);
-	term_room(r1 < xtop ? xtop - xrow : r1 - r2);
+	term_room(r1 < xtop ? xtop - xrow : r1 - r2 - !!*vi_word);
 	xrow = r1;
 	if (r1 < xtop)
 		xtop = r1;
 	sbuf_mem(sb, ln, l1)
-	led_input(sb, &post, postn, r1 - (r1 - r2), 0);
-	sbufn_str(sb, post)
-	if (sb->s_n != tlen || memcmp(ln + l1, sb->s + l1, tlen - l2 - l1))
+	key = led_input(sb, post, postn, r1 - (r1 - r2), 0, &postn);
+	if (postn + l2 != tlen || memcmp(ln + l1, sb->s + l1, tlen - l2 - l1))
 		lbuf_edit(xb, sb->s, r1, r2 + 1, o1, xoff);
 	free(sb->s);
-	free(_post);
+	return key;
 }
 
 static void vi_case(int r1, int o1, int r2, int o2, int lnmode, int cmd)
@@ -964,7 +963,7 @@ static void vi_shift(int r1, int r2, int dir, int count)
 	free(sb->s);
 }
 
-static void vc_motion(int cmd)
+static int vc_motion(int cmd)
 {
 	int r1 = xrow, r2 = xrow;	/* region rows */
 	int o1 = xoff, o2;		/* visual region columns */
@@ -978,9 +977,9 @@ static void vc_motion(int cmd)
 	if ((mv = vi_motionln(&r2, cmd, vi_arg ? vi_arg : 1)))
 		o2 = -1;
 	else if (!(mv = vi_motion(1, &r2, &o2)))
-		return;
+		return 0;
 	if (mv < 0)
-		return;
+		return 0;
 	lnmode = o2 < 0;
 	if (lnmode) {
 		o1 = 0;
@@ -999,13 +998,13 @@ static void vc_motion(int cmd)
 			o2++;
 	if (cmd == 'y') {
 		vi_yank(r1, o1, r2, o2, lnmode);
-		return;
+		return 0;
 	}
 	mv = lbuf_len(xb);
 	if (cmd == 'd')
 		vi_delete(r1, o1, r2, o2, lnmode);
 	else if (cmd == 'c')
-		vi_change(r1, o1, r2, o2, lnmode);
+		return vi_change(r1, o1, r2, o2, lnmode);
 	else if (cmd == '~' || cmd == 'u' || cmd == 'U')
 		vi_case(r1, o1, r2, o2, lnmode, cmd);
 	else if (cmd == '!')
@@ -1016,12 +1015,13 @@ static void vc_motion(int cmd)
 	else if (cmd == TK_CTL('w'))
 		vi_shift(r1, r2, -1, INT_MAX / 2);
 	vi_mod |= r1 != r2 || mv != lbuf_len(xb) ? 1 : 2;
+	return 0;
 }
 
-static void vc_insert(int cmd)
+static int vc_insert(int cmd)
 {
 	char *post, *ln = lbuf_get(xb, xrow);
-	int row, cmdo, l1, off, postn = 1;
+	int row, cmdo, l1, off, key, postn = 1;
 	sbuf_smake(sb, xcols)
 	if (cmd == 'I')
 		xoff = lbuf_indents(xb, xrow);
@@ -1054,12 +1054,11 @@ static void vc_insert(int cmd)
 	term_pos(row - xtop, 0);
 	term_room(cmdo);
 	sbuf_mem(sb, ln, l1)
-	led_input(sb, &post, postn, row, cmdo);
-	if (sb->s_n != l1 || cmdo || !ln) {
-		sbufn_str(sb, post)
+	key = led_input(sb, post, postn, row, cmdo << 2, &postn);
+	if (postn != l1 || cmdo || !ln)
 		lbuf_edit(xb, sb->s, row, row + !cmdo, off, xoff);
-	}
 	free(sb->s);
+	return key;
 }
 
 static int vc_put(int cmd)
@@ -1100,31 +1099,10 @@ static int vc_put(int cmd)
 
 static void vc_join(int spc, int cnt)
 {
-	int beg = xrow;
-	int end = xrow + cnt;
-	int off = 0;
-	if (!lbuf_get(xb, beg) || !lbuf_get(xb, end - 1))
+	int o2 = 0;
+	if (lbuf_join(xb, xrow, xrow + cnt, xoff, &o2, spc))
 		return;
-	sbuf_smake(sb, 1024)
-	for (int i = beg; i < end; i++) {
-		char *ln = lbuf_get(xb, i);
-		char *lnend = ln + lbuf_s(ln)->len;
-		if (i > beg) {
-			while (ln[0] == ' ' || ln[0] == '\t')
-				ln++;
-			if (spc && sb->s_n && *ln != ')' &&
-					sb->s[sb->s_n-1] != ' ') {
-				sbuf_chr(sb, ' ')
-				off++;
-			}
-		}
-		off += (i+1 == end) ? 0 : uc_slen(ln) - 1;
-		sbuf_mem(sb, ln, lnend - ln)
-	}
-	sbufn_chr(sb, '\n')
-	lbuf_edit(xb, sb->s, beg, end, xoff, off);
-	xoff = off;
-	free(sb->s);
+	xoff = o2;
 	vi_mod |= 1;
 }
 
@@ -1200,13 +1178,23 @@ static void vi_argcmd(int arg, char cmd)
 	term_push(str, cs - str + 1);
 }
 
+#define topfix() \
+if (xrow < 0 || xrow >= lbuf_len(xb)) \
+	xrow = lbuf_len(xb) ? lbuf_len(xb) - 1 : 0; \
+if (xtop > xrow) \
+	xtop = xtop - xrows / 2 > xrow ? \
+			MAX(0, xrow - xrows / 2) : xrow; \
+if (xtop + xrows <= xrow) \
+	xtop = xtop + xrows + xrows / 2 <= xrow ? \
+			xrow - xrows / 2 : xrow - xrows + 1; \
+
 void vi(int init)
 {
 	char *ln, *cs;
 	int mv, n, k, c;
 	xgrec++;
 	if (init) {
-		xtop = MAX(0, xrow - xrows / 2);
+		topfix()
 		vi_col = vi_off2col(xb, xrow, xoff);
 		vi_drawagain(xtop);
 		vi_drawmsg();
@@ -1313,7 +1301,7 @@ void vi(int init)
 			case TK_CTL('n'):
 				vi_cndir = vi_arg ? -vi_cndir : vi_cndir;
 				vi_arg = ex_buf - bufs + vi_cndir;
-			case TK_CTL('_'): /* note: this is also ^7 per ascii */
+			case TK_CTL('_'):	/* this is also ^7 on some systems */
 				if (vi_arg > 0)
 					goto switchbuf;
 				xleft = 0;
@@ -1361,7 +1349,7 @@ void vi(int init)
 				break;
 			case TK_CTL('k'):;
 				static struct lbuf *writexb;
-				if ((cs = ex_exec("w")) && xb == writexb)
+				if ((cs = ex_exec("w")) && writexb && xb == writexb)
 					cs = ex_exec("mpt0:w!");
 				writexb = cs ? xb : NULL;
 				break;
@@ -1382,9 +1370,6 @@ void vi(int init)
 						term_push("j", 1);
 						term_push(rep_cmd, rep_len);
 						if (strchr("iIoOaAsScC", rep_cmd[0])) {
-							/* go to the left to restore
-							previous position of what
-							was inserted. */
 							term_push("0", 1);
 							if (noff)
 								vi_argcmd(noff, 'l');
@@ -1392,9 +1377,6 @@ void vi(int init)
 						vi_arg--;
 					}
 					break;
-				case 'j':
-					vi_joinmode = !vi_joinmode;
-					break;
 				case 'w':
 					vi_nlword = !vi_nlword;
 					break;
@@ -1512,7 +1494,7 @@ void vi(int init)
 			case '>':
 			case '<':
 			case TK_CTL('w'):
-				vc_motion(c);
+				k = vc_motion(c);
 				if (c == 'c')
 					goto ins;
 				break;
@@ -1522,10 +1504,10 @@ void vi(int init)
 			case 'A':
 			case 'o':
 			case 'O':
-				vc_insert(c);
+				k = vc_insert(c);
 				ins:
 				vi_mod |= !xpac && xrow == orow ? 8 : 1;
-				if (vi_insmov == 127) {
+				if (k == 127) {
 					if (xrow && !(xoff > 0 && lbuf_eol(xb, xrow, 1))) {
 						xrow--;
 						vc_join(0, 2);
@@ -1538,7 +1520,7 @@ void vi(int init)
 					xoff--;
 				break;
 			case 'J':
-				vc_join(vi_joinmode, vi_arg <= 1 ? 2 : vi_arg);
+				vc_join(1, vi_arg <= 1 ? 2 : vi_arg);
 				break;
 			case 'K': {
 				preserve(int, xled, xled = 0;)
@@ -1561,8 +1543,7 @@ void vi(int init)
 				vi_mod |= 1;
 				break;
 			case 'm':
-				if ((k = term_read()) > 0 && islower(k))
-					lbuf_mark(xb, k, xrow, xoff);
+				lbuf_mark(xb, term_read(), xrow, xoff);
 				break;
 			case 'p':
 			case 'P':
@@ -1708,14 +1689,7 @@ void vi(int init)
 				rep_len = icmd_pos;
 			}
 		}
-		if (xrow < 0 || xrow >= lbuf_len(xb))
-			xrow = lbuf_len(xb) ? lbuf_len(xb) - 1 : 0;
-		if (xtop > xrow)
-			xtop = xtop - xrows / 2 > xrow ?
-					MAX(0, xrow - xrows / 2) : xrow;
-		if (xtop + xrows <= xrow)
-			xtop = xtop + xrows + xrows / 2 <= xrow ?
-					xrow - xrows / 2 : xrow - xrows + 1;
+		topfix()
 		ln = lbuf_get(xb, xrow);
 		xoff = ren_noeol(ln, xoff);
 		if (ln && !rstate->wid[xoff]) {
@@ -1737,11 +1711,11 @@ void vi(int init)
 			static char *word;
 			if ((cs = vi_curword(xb, xrow, xoff, xhlw))) {
 				if (!word || strcmp(word, cs)) {
-					syn_reloadft(syn_addhl(cs, 1));
-					free(word);
-					word = cs;
+					syn_reloadft(syn_addhl(cs, 1), 0);
 					vi_mod |= 1;
 				}
+				free(word);
+				word = cs;
 			}
 		}
 		if (xhlp && (k = syn_findhl(3)) >= 0) {
@@ -1780,9 +1754,9 @@ void vi(int init)
 				if (!(vi_mod & 1) && !*vi_word)
 					vi_drawrow(orow);
 			syn_blockhl = -1;
-			syn_reloadft(syn_addhl("^.+", 2));
+			syn_reloadft(syn_addhl("^.+", 2), 0);
 			vi_drawrow(xrow);
-			syn_reloadft(syn_addhl(NULL, 2));
+			syn_reloadft(syn_addhl(NULL, 2), 0);
 		} else if (vi_mod & 2 && !(vi_mod & 1)) {
 			syn_blockhl = -1;
 			vi_drawrow(xrow);
@@ -1842,7 +1816,7 @@ int main(int argc, char *argv[])
 				xvis &= ~4;
 			else {
 				fprintf(stderr, "Unknown option: -%c\n", argv[i][j]);
-				fprintf(stderr, "Nextvi-2.3 Usage: %s [-emsv] [file ...]\n", argv[0]);
+				fprintf(stderr, "Nextvi-3.2 Usage: %s [-emsv] [file ...]\n", argv[0]);
 				return EXIT_FAILURE;
 			}
 		}
